import logging
import os
import re
import time

from enum import Enum
from urllib.request import urlopen

class Gutenberg_Dataset():
    """ A fuzzy, lightweight class to access, search and filter Project Gutenberg resources
    
    GutenbergLib by default uses a mirror's root URL. Alternatively, you can specify a local directory containing a Gutenberg mirror.
    That mirror directory needs to contain a GUTINDEX.ALL file and has typically many 
    sub-directories `0` ,.. `n` .

    A mirror of project Gutenberg can be created by:
    
    .. code-block:: console

        #!/bin/bash
        rsync -zarv --dry-run --prune-empty-dirs --del --include="*/" --include='*.'{txt,pdf,ALL} --exclude="*" aleph.gutenberg.org::gutenberg ./gutenberg_mirror

    You can remove the PDF files, since they are currently not used, and need to review the `--dry-run` option.

    Note: :func:`~Gutenberg_Dataset.Gutenberg_Dataset.load_index` needs to be called before any other methods.
    
    :param root_url: url of Project Gutenberg or any mirror URL, or a local directory containing a Gutenberg mirror.
    :param cache_dir: path to a directory that will be used to cache the Gutenberg index and already downloaded texts. The cache directory is only used, if a remote Gutenberg URL and not a local mirror is used.
    """
    def __init__(self, root_url="http://www.mirrorservice.org/sites/ftp.ibiblio.org/pub/docs/books/gutenberg", cache_dir="gutenberg"):
        self.log = logging.getLogger('GutenbergLib')
        self.root_url = root_url
        self.index=None
        self.NEAR=2048
        self.start_tokens=["*** START OF THIS PROJECT", "E-text prepared by", 
                           "This book was generously provided by the ", "*** START OF THIS PROJECT GUTENBERG",
                           "START OF THE PROJECT GUTENBERG"]    
        self.near_start_tokens=["produced by ", "Produced by ", ", prepared by", "Transcriber's Note", 
                                "Transcriber's note:", "Anmerkungen zur Tanskription", 
                                "Distributed Proofreading Team", "offensichtliche Schreibfehler",
                                "Inkonsistenzen in der Rechtschreibung", "Im Original",
                                "Obvious printer errors", "spelling was kept",
                                "ERRATA have been applied", "punctuation errors", "have been silently corrected",
                                "changes to the text", "Transcriber note", "Transcriber Note","_italic_",
                                "Variable spelling"] 
        self.end_tokens=["End of the Project Gutenberg", "*** END OF THIS PROJECT", 
                         "***END OF THE PROJECT GUTENBER", "Ende dieses Projekt Gutenberg", 
                         "*** END OF THE PROJECT GUTENBERG",
                         "End of Project Gutenberg", "Transcriber's Note"]
        self.local_mirror=False
        if root_url[:4]!='http':
            if not os.path.exists(root_url):
                self.log.error(f'If root_url points to non-http URL, it must be an existing local directory containing a Gutenberg mirror: {root_url}')
            else:
                index_path=os.path.join(root_url, "GUTINDEX.ALL")
                if not os.path.exists(index_path):
                    self.log.error(f'GUTINDEX.ALL not found in {root_url}, this is not a valid Gutenberg mirror')
                else:
                    self.local_mirror=True
                    self.cache_dir=None
            return
        if self.local_mirror is False:
            try:
                if not os.path.exists(cache_dir):
                    os.makedirs(cache_dir)
                self.cache_dir=cache_dir
            except Exception as e:
                self.cache_dir=None
                self.log.error(f"Failed to create cache directory {cache_dir}, {e}")

    def _parse_record(self,record,verbose=True):
        """ internal function to recreate some consistent record information from near-freestyle text """
        rl=record.split('\n')
        white=str(chr(160))+str(chr(9))+" " # non-breaking space, TAB, and space
        ebook_no=""
        while len(rl[0])>0 and rl[0][-1] in white:
            rl[0]=rl[0][:-1]
        while len(rl[0])>0 and not rl[0][-1] in white:
            ebook_no=rl[0][-1]+ebook_no
            rl[0]=rl[0][:-1]
        while len(rl[0])>0 and rl[0][-1] in white:
            rl[0]=rl[0][:-1]
        
        # Sanity check
        try:
            fa=re.findall(ebook_no,"\A[0-9]+[A-C]\Z")
        except Exception as e:
            fa=None
            if verbose is True:
                self.log.warning(f"Failed to apply regex on >{ebook_no}<")
            
        if len(rl[0])<5 or fa==None or len(ebook_no)>7:
            if verbose is True:
                print("-------------------------------------")
                print(record)
                print("- - - - - - - - - - - - - - - - - - -")
                print(f"Dodgy record: {rl[0]}")
                print(f"    ebook-id:  >{ebook_no}<")
            return None
        
        for i in range(len(rl)):
            rl[i]=rl[i].strip()
            
        p=0
        while p<len(rl)-1:
            if len(rl[p+1])==0:
                print(f"Invalid rec: {record}")
                p+=1
            else:
                if rl[p+1][0]!="[":
                    rl[p]+=" "+rl[p+1]
                    del rl[p+1]
                    if rl[p][-1]==']':
                        p+=1
                else:
                    p+=1
        
        rec={}
        l0=rl[0].split(", by ")
        rec['title']=l0[0]
        rec['ebook_id']=ebook_no
        # if len(l0)>2:
        #    print(f"Chaos title: {rl[0]}")
        if len(l0)>1:
            rec['author']=l0[-1]
        for r in rl[1:]:
            if r[0]!='[' or r[-1]!=']':
                if r[0]=='[':
                    ind=r.rfind(']')
                    if ind != -1:
                        # print(f"Garbage trail {r}")
                        r=r[:ind+1]
                        # print(f"Fixed: {r}")
                    else:
                        # print(f"Missing closing ] {r}")
                        r+=']'
                        # print(f"Fixed: {r}")
            if r[0]=='[' and r[-1]==']':
                r=r[1:-1]
                i1=r.find(':')
                if i1==-1:
                    r=r.replace("Author a.k.a.","Author a.k.a.:")
                    i1=r.find(':')
                if i1!=-1:
                    i2=r[i1:].find(' ')+i1
                else:
                    i2=-1
                if i1==-1 and i2==-1:
                    pass
                    # print(f"Invalid attribut in {rl}::{r}")
                else:
                    if i2-i1==1:
                        key=r[:i1]
                        val=r[i2+1:]
                        if '[' in key or ']' in key or '[' in val or ']' in val or len(key)>15:
                            pass
                            # print("messy key/val")
                        else:
                            rec[key.strip().lower()]=val.strip()
                    else:
                        pass
                        # print(f"Bad attribute name terminator, missing ': ' {r}")
            else:
                pass
                # print(f"Invalid attribut in {rl}::{r}")
        if len(rec)>1:
            if "language" not in rec.keys():
                rec["language"]="English"
        return rec
        
    def _parse_index(self, lines):
        """ internal function to parse the fuzzy text-based Gutenberg table of content """
        class State(Enum):
            NONE=1,
            SYNC_START=2,
            SYNC_REC=3,
            END=5
    
        white=str(chr(160))+str(chr(9))+" " # non-breaking space, TAB, and space
        state=State.NONE
        start_token="~ ~ ~ ~"
        stop_token=["====="]
        end_token="<==End"
        ignore_headers=["TITLE and AUTHOR"]
        ignore_content=["Not in the Posted Archives","human-read audio ebooks", "Audio:"]
        empty_lines=0
        records=[]
        for line in lines:
            if line[:len(end_token)]==end_token:
                state=State.END
                break

            if state==State.NONE:
                if line[:len(start_token)]==start_token:
                    state=State.SYNC_START
                    empty_lines=0
                    continue
            if state==State.SYNC_START:
                if len(line.strip())==0:
                    empty_lines+=1
                    if empty_lines>1:
                        state=State.NONE
                        continue
                else:
                    stopped=False
                    for stop in stop_token:
                        if line[:len(stop)]==stop:
                            stopped=True
                            break
                    if stopped is True:
                        state=State.NONE
                        empty_lines=0
                        continue
                    ignore=False
                    for header in ignore_headers:
                        if line[:len(header)]==header:
                            empty_lines=0
                            ignore=True
                    for token in ignore_content:
                        if token in line:
                            empty_lines=0
                            ignore=True
                    if ignore is True:
                        continue
                    rec=line
                    state=State.SYNC_REC
                    continue
            if state==State.SYNC_REC:
                if len(line.strip())==0 or line[0] not in white:
                    if len(records)<10:
                        parsed_rec=self._parse_record(rec, verbose=True)
                    else:
                        parsed_rec=self._parse_record(rec, verbose=False)
                        
                    if parsed_rec is not None:
                        records.append(parsed_rec)
                    empty_lines=1
                    if len(line.strip())==0:
                        state=State.SYNC_START
                        continue
                    else:
                        rec=line
                        continue
                rec=rec+"\n"+line
        return records
                    
    def load_index(self, cache=True, cache_expire_days=30):
        """ This function loads the Gutenberg record index, either from cache, or from a website
        
        This should be the first method being used, since many other methods rely on the index being loaded.

        :param cache: default `True`, use the cache directory to cache both index and text files. Index expires after `cache_expire_days`, text files never expire. Should *NOT* be set to `False` in order to prevent unnecessary re-downloading.
        :param cache_expire_days: Number of days after which the index is re-downloaded.
        """
        raw_index=None
        if self.local_mirror is False:
            if self.cache_dir is None:
                self.log.error("Cannot cache library index, no valid cache directory.")
                return False
            ts_file=os.path.join(self.cache_dir,"timestamp")
            cache_file=os.path.join(self.cache_dir,"gutenberg_index")
            expired=True
            read_from_cache=False
            if os.path.isfile(ts_file) and os.path.isfile(cache_file):
                try:
                    with open(ts_file,'r') as f:
                        ts=float(f.read())
                    if time.time()-ts<cache_expire_days*24*3600:
                        expired=False
                        read_from_cache = True
                        self.log.debug("Cache timestamp read.")
                    else:
                        self.log.info("Cache for Gutenberg-index is expired, reloading from web.")
                except:
                    self.log.warning("Failed to read cache timestamp, reloading from web.")
            if expired is False and os.path.isfile(cache_file):
                try:
                    with open(cache_file,'r') as f:
                        raw_index=f.read()
                        self.log.info(f"Gutenberg index read from local cache: {cache_file}")
                except:
                    expired=True
                    self.log.warning("Failed to read cached index, reloading from web.")
            if expired is True:
                index_url=self.root_url+"/GUTINDEX.ALL"
                try:
                    raw_index = urlopen(index_url).read().decode('utf-8')
                    if raw_index[0]=='\ufeff':  # Ignore BOM
                        raw_index=raw_index[1:]
                    raw_index=raw_index.replace('\r','')
                    self.log.info(f"Gutenberg index read from {index_url}")
                except Exception as e:
                    self.log.error(f"Failed to download Gutenberg index from {index_url}, {e}")
                    return False
            if cache is True and read_from_cache is False:
                try:
                    with open(ts_file,'w') as f:
                        f.write(str(time.time()))
                        self.log.debug("Wrote read cache timestamp.")
                except Exception as e:
                    self.log.error(f"Failed to write cache timestamp to {ts_file}, {e}")
                try:
                    with open(cache_file,'w') as f:
                        f.write(raw_index)
                        self.log.debug("Wrote read cached index.")
                except Exception as e:
                    self.log.error(f"Failed to write cached index to {cache_file}, {e}")
        else:
            index_file=os.path.join(self.root_url, "GUTINDEX.ALL")
            try:
                with open(index_file,'r') as f:
                    raw_index=f.read()
                    if raw_index[0]=='\ufeff':  # Ignore BOM
                        raw_index=raw_index[1:]
                    raw_index=raw_index.replace('\r','')
                    self.log.info(f"Gutenberg index read from local mirror: {index_file}")
            except Exception as e:
                self.log.error(f"Failed to read Gutenberg index from local mirror: {index_file}, {e}")
                return
        lines=raw_index.split('\n')
        self.records=self._parse_index(lines)

    def load_book(self, ebook_id):
        """ get text of an ebook from Gutenberg by ebook_id 
        
        This function returns the unfiltered raw text including all Gutenberg headers and footers.
        Use :func:`~Gutenberg_Dataset.Gutenberg_Dataset.get_book` to retrieve a dictionary with metadata and filtered text.

        :param ebook_id: Gutenberg id (Note: string, since this sometimes contains a character!)
        :returns: book text as string, unfiltered. Can be filtered with :func:`~Gutenberg_Dataset.Gutenberg_Dataset.filter_text`
        """
        txt, dl, val = self._load_book_ex(ebook_id)[0]
        if val is True:
            return txt, dl
        else:
            return None, dl

    def _read_download(self, filenames, path_stub, cache_name):
        """ Internal function to read ebook from cache or download it. """
        cache_file=None
        downloaded=False
        if self.cache_dir is not None:
            cache_file=os.path.join(self.cache_dir,cache_name)
            if os.path.isfile(cache_file):
                try:
                    with open(cache_file,'r') as f:
                        data=f.read()
                        self.log.info(f"Book read from cache at {cache_file}")
                        downloaded = False
                        return data, None, downloaded
                except Exception as e:
                    self.log.error(f"Failed to read cached file {cache_file}")
        data=None
        file_url=None
        for filename, encoding in filenames:
            file_url=self.root_url+path_stub+filename
            if self.local_mirror is False:
                try:
                    if encoding != 'bin':
                        data = urlopen(file_url).read().decode(encoding)
                    else:
                        data = urlopen(file_url).read()
                    self.log.info(f"Book read from {file_url}")
                    downloaded = True
                    break
                except Exception as e:
                    self.log.debug(f"Failed to download {file_url}, {e}")
            else:
                try:
                    if encoding != 'bin':
                        with open(file_url,'r',encoding=encoding) as f:
                            data = f.read()
                    else:
                        with open(file_url,'rb') as f:
                            data = f.read()
                    self.log.info(f"Book read from local mirror at {file_url}")
                    downloaded = False
                    break
                except Exception as e:
                    self.log.debug(f"Failed to read local mirror at {file_url}, {e}")
        return data, cache_file, downloaded

    def _load_book_ex(self, ebook_id):
        """ Internal function to get text of an ebook from Gutenberg by ebook_id with remote download information
        
        This function returns the unfiltered raw text including all Gutenberg headers and footers and a boolean flag
        indicating with 'True', if the book was downloaded from a remote source, 'False' indicates cached book retrieval
        without remote access.

        The flag can be used to control or limit the number of books downloaded from the remote source.

        Use :func:`~Gutenberg_Dataset.Gutenberg_Dataset.get_book` to retrieve a dictionary with metadata and filtered text.

        :param ebook_id: Gutenberg id (Note: string, since this sometimes contains a character!)
        :returns: tuple: book text as string, unfiltered, and a flag indicating with 'True' if book was downloaded from remote, validity flag, 'True' indicates valid text. 
        """
        if ebook_id is None or len(ebook_id)==0:
            return None
        if ebook_id[-1]=='C':
            ebook_id=ebook_id[:-1]
        path_stub=""
        downloaded = False
        valid = False
        
        for i in range(len(ebook_id)-1):
            path_stub+="/"+ebook_id[i]
        path_stub+="/"+ebook_id+"/"
        filenames=[(ebook_id+"-0.txt",'utf-8'), (ebook_id+".txt",'utf-8'), (ebook_id+"-8.txt","latin1"), (ebook_id+".txt","latin1")]
        cache_name=ebook_id+".txt"
        data, cache_file, downloaded = self._read_download(filenames, path_stub, cache_name)
        if data is not None:
            if data[0]=='\ufeff':  # Ignore BOM
                data=data[1:]
            data=data.replace('\r','')
            valid = True
        else:
            filenames = [(ebook_id+"-pdf.pdf","bin")]
            cache_name=ebook_id+".pdf"
            data, cache_file, downloaded = self._read_download(filenames, path_stub, cache_name)
            if data is not None:
                self.log.error(f"Ebook {cache_name} is only available in PDF format, this is not supported.")
            else:
                self.log.warning(f"Failed to download {filenames}, skipping book.")
                return None, downloaded, False
        if cache_file is not None:
            try:
                with open(cache_file,'w') as f:
                    f.write(data)
            except:
                self.log.error(f"Failed to cache file {cache_file}")
        return data, downloaded, valid
    
    def filter_text(self, book_text, add_start_tokens=None, add_near_start_tokens=None, add_end_tokens=None):
        """ Heuristically remove header and trailer texts not part of the actual books

        Unfortunatelly, formatting of Gutenberg books is an unbelievable mess. Using lists of tokens `self.start_tokens` (indicating
        the start of the actual book text), `self.near_start_tokens` (indicating possibly ambiguous tokens near a `start_tokens` token, 
        further narrowing the start of text), and `self.end_tokens` (indicating the end of the book text), this function tries to find 
        the start and end of the book text. The user can either extend the lists of class member tokens, of provide temporary additional 
        tokens as parameter to this function.
        
        The list of `start_tokens` contains only tokens that are always significant as being part of header-cruft (e.g. 'START OF THIS GUTENBERG'). 
        `near_start_tokens` are tokens that might be ambiguous, but are still part of the header-cruft, (e.g. 'produced by'). 
        `near_start_tokens` are only used, if they are within `self.NEAR` bytes to the latest `start_tokens` token, 
        to heuristically prevent false positives.

        *Note:* Use logging via `logging.basicConfig(level=logging.DEBUG)` to analyze the filtering process.

        :param book_text: text of the book (string)
        :param add_start_tokens: additional start tokens (list of strings)
        :param add_near_start_tokens: additional near start tokens (list of strings)
        :param add_end_tokens: additional end tokens (list of strings)
        :returns: filtered text (string)
        """
        start_tokens = self.start_tokens
        if add_start_tokens is not None:
            start_tokens.extend(add_start_tokens)
        near_start_tokens = self.near_start_tokens
        if add_near_start_tokens is not None:
            near_start_tokens.extend(add_near_start_tokens)
        end_tokens = self.end_tokens
        if add_end_tokens is not None:
            end_tokens.extend(add_end_tokens)

        if book_text is None:
            self.log.warning("Filter: book text is None, returning None")
            return None
        blen=len(book_text)
        
        pstart=0
        for token in start_tokens:
            pos=book_text.find(token)
            if pos > pstart:
                pstart = pos
                self.log.debug(f"Start-token [{token}] found at position {pos}")
        if pstart>0:
            pos=book_text[pstart:].find("\n\n")
            if pos>=0 and pos <= self.NEAR:
                pos += pstart
                while book_text[pos]=='\n':
                    pos += 1  # eof?!
                pstart=pos
        if pstart>blen/2:
            self.log.warning("Preamble is taking more than half of the book!")
        new_book=book_text[pstart:]
        xpos=-1
        for token in near_start_tokens:
            pos=new_book.find(token)
            if pos>=0 and pos<=self.NEAR:
                self.log.debug(f"Near-Start-token [{token}] found at position {pos}")
                if pos>xpos:
                    xpos=pos
        if xpos > -1:
            pos2=new_book[xpos:].find("\n\n")
            if pos2<=self.NEAR and pos2>0:
                self.log.debug(f"Trying extra skipping (2) for {pos2}...")
                while new_book[xpos+pos2]=='\n':
                    pos2 += 1
                new_book=new_book[xpos+pos2:]
                self.log.debug(f"Additionally shortened start by {xpos+pos2} chars")
            else:
                pos2=new_book[xpos:].find("\n")
                if pos2<=self.NEAR and pos2>0:
                    self.log.debug(f"Trying extra skipping (3) for {pos2}...")
                    while new_book[xpos+pos2]=='\n':
                        pos2 += 1
                    new_book=new_book[xpos+pos2:]
                    self.log.debug(f"Additionally shortened start by {xpos+pos2}, {xpos}+{pos2} chars")
                else:
                    pos2=0
                    new_book=new_book[xpos+pos2:]
        
        pend=len(new_book)
        for token in end_tokens:
            pos=new_book.find(token)
            if pos!=-1 and pos < pend:
                self.log.debug(f"End-token [{token}] found at pos {pos}")
                pend = pos
        if pend<len(new_book):
            pos=new_book[:pend].rfind("\n\n")
            if pos>0:
                while new_book[pos]=='\n':
                    pos -= 1  # eof?!
                pend=pos+1
        else:
            self.log.debug("No end token found!")
        if pend<len(new_book)/2:
            self.log.debug("End-text is taking more than half of the book!")
        new_book=new_book[:pend]
        return new_book
        
    def find_keywords(self,*search_keys):
        """ Search of an arbitrary number of keywords in a book record
       
        *Note:* :func:`~Gutenberg_Dataset.Gutenberg_Dataset.load_index` needs to be called once before this function can be used.

        :returns: list of records that contain all keywords in any field.
        """
        frecs=[]
        for rec in self.records:
            found=True
            for sk in search_keys:
                subkey=False
                for key in rec.keys():
                    if sk.lower() in key.lower() or sk.lower() in rec[key].lower():
                        subkey=True
                        break
                if subkey is False:
                    found=False
                    break
            if found is True:
                frecs += [rec]
        return frecs
    
    def search(self, search_dict):
        """ Search for book record with key specific key values
        For a list of valid keys, use `get_record_keys()`
        Standard keys are: `ebook_id`, `author`, `language`, `title`

        *Note:* :func:`~Gutenberg_Dataset.Gutenberg_Dataset.load_index` needs to be called once before this function can be used.

        Example: `search({"title": ["philosoph","phenomen","physic","hermeneu","logic"], "language":"english"})`
        Find all books whose titles contain at least one of the keywords, language english. Search keys can either be
        search for a single keyword (e.g. english), or an array of keywords. 
        
        :returns: list of records """
        if not hasattr(self, 'records') or self.records is None:
            self.log.debug("Index not loaded, trying to load...")
            self.load_index()
        frecs=[]
        for rec in self.records:
            found=True
            for sk in search_dict:
                if sk not in rec:
                    found=False
                    break
                else:
                    skl=search_dict[sk]
                    if not isinstance(skl,list):
                        skl=[skl]
                    nf=0
                    for skli in skl:
                        if skli.lower() in rec[sk].lower():
                            nf=nf+1
                    if nf==0:
                        found=False
                        break
            if found is True:
                frecs += [rec]
        return frecs
        
    def insert_book_texts(self, search_dict, download_count_limit=20):
        """ Inserts book texts into the records returned by :func:`~Gutenberg_Dataset.Gutenberg_Dataset.search`.

        In order to prevent the download of too many books, the download count limit is set to `download_count_limit`.
        Downloaded books are cached and cached books are not counted towards the download count limit. Calling this
        function again will download books that have not been downloaded yet. The filtered book content is inserted
        into the dictionary with the key `text`.
        
        :param search_dict: search array of dictionaries that at least contain the key `ebook_id`.
        :param download_count_limit: maximum number of books to download, if no local mirror is used. No limits apply for local mirrors.
        :returns: list of records including filtered book text-based in the `text` field.
        """
        dls=0
        for i in range(0, len(search_dict)):
            self.log.debug(f"Getting id={search_dict[i]['ebook_id']}, {search_dict[i]['title']}")
            bt, dl, val = self._load_book_ex(search_dict[i]["ebook_id"])
            if bt is None or val is False:
                if val is False:
                    self.log.warning(f"Download of book {search_dict[i]['ebook_id']}, {search_dict[i]['title']}: invalid format!")
                else:
                    self.log.error(f"Download of book {search_dict[i]['ebook_id']}, {search_dict[i]['title']} failed!")
                continue
            search_dict[i]['text']=self.filter_text(bt)
            if dl is True and self.local_mirror is False:
                dls += 1
                if dls>download_count_limit:
                    self.log.error(f"Download limit reached ({download_count_limit}), stopping download...")
                    break
        return search_dict  

    def get_book(self, ebook_id: str):
        """ Get a book record metadata and filtered text by its ebook_id
       
        This function returns a dictionary with metadata and filtered text. Use :func:`~Gutenberg_Dataset.Gutenberg_Dataset.load_book` 
        to get the raw unfiltered text.
        
        *Note:* :func:`~Gutenberg_Dataset.Gutenberg_Dataset.load_index` needs to be called once before this function can be used.

        :param ebook_id: ebook_id (String, since some IDs contain letters) of the book to be retrieved
        :returns: book record (dictionary with metadata and filtered text)
        """
        for rec in self.records:
            if rec["ebook_id"]==ebook_id:
                rec['text']=self.filter_text(self.load_book(ebook_id))
                return rec
        return None

    def get_record_keys(self):
        """ Get a list of all keys that are used within records. 
        Standard keys are: `ebook_id`, `author`, `language`, `title`.
        
        *Note:* :func:`~Gutenberg_Dataset.Gutenberg_Dataset.load_index` needs to be called once before this function can be used.

        :returns: list of all different keys that are somehow used."""
        rks=[]
        for r in self.records:
            rks=set(list(rks) + list(r.keys()))
        return rks

    def get_unique_record_values(self, key):
        """ Get a list of all unique values a given keys has for all records.
        
        *Note:* :func:`~Gutenberg_Dataset.Gutenberg_Dataset.load_index` needs to be called once before this function can be used.

        Example: `get_unique_records_values('language')` returns all languages in Gutenberg.

        :param key: key to search for.
        :returns: list of all unique values for a given key.
        """
        uv=[]
        if key not in self.get_record_keys():
            print(f"{key} is not a key used in any record!")
            return None
        for r in self.records:
            if key in r:
                uv=set(list(uv)+[r[key]])
        uv=sorted(uv)
        return uv



import random
import logging
try:
   from IPython.core.display import display, HTML
except:
    pass

class Text_Dataset:
    """ Initialize the Text_Dataset with a list of texts.

    The Gutenberg_Dataset can be used to create such a list, by:
        
    .. code-block:: python

        from ml_indie_tools.Gutenberg_Dataset import Gutenberg_Dataset
        from ml_indie_tools.Text_Dataset import Text_Dataset
        gd = Gutenberg_Dataset()
        gd.load_index()
        ls = gd.search({'author': 'kant', 'title': 'kritik', 'language': 'german'})  # returns a list of texts
        ls = gd.insert_texts(ls)  # this inserts the actual text of the books into field 'text'.
        # Now ls contains a valid list of text records:
        td = Text_Dataset(ls)
    
    :param text_list: list of text-records of the form: {'author': 'author', 'title': 'title', 'language': 'some-language', 'text': 'the-long-text'}. Optinal parameters: 'weight': 1.0
    :param sanitize_white_space: If True, white space is replaced by a single space.
    :param separate_punctuation: If True, punctuation is separated from words.
    :param preserve_case: If True, the case of the text is preserved.
    """
    def __init__(self, text_list, sanitize_white_space=False, separate_punctuation=False, preserve_case=True):
        self.log = logging.getLogger("Datasets")
        self.text_list = []
        self.index = 1
        self.word_tokenizer_init = False
        self.char_tokenizer_init = False
        self.getitem_init = False
        
        req_attrs=['title', 'author', 'language', 'text']
        for ind in range(0,len(text_list)):
            valid=True
            miss=[]
            for attr in req_attrs:
                if attr not in text_list[ind]:
                    valid=False
                    miss.append(attr)
            if valid is False:
                self.log.error(f"Missing attribute(s) {miss} in text[{ind}], skipping")
                continue
            text=text_list[ind]
            text['index']=self.index
            text['text']=self.filter_text(text['text'], sanitize_white_space=sanitize_white_space, separate_punctuation=separate_punctuation, preserve_case=preserve_case)
            self.index += 1
            self.text_list.append(text)
        self.log.info(f"Loaded {len(self.text_list)} texts")
        self._calc_probability_weights()

    def _calc_probability_weights(self):
        prs = 0
        for text in self.text_list:
            if 'weight' in text:
                w=text['weight']
            else:
                w=1.0
            pr = len(text['text']) * w
            prs = prs + pr
            text['probability_weight'] = pr
        for text in self.text_list:
            text['probability_weight'] = text['probability_weight'] / prs
        self.tidx = []
        self.tcum = []
        tc=0
        for idx in range(0,len(self.text_list)):
            self.tidx.append(idx)
            text = self.text_list[idx]
            self.tcum.append(text['probability_weight'] + tc)
            tc = self.tcum[-1]

    def _get_random_text_index(self, weighted=True):
        """ Return a random text index from the Text_Dataset.
        
        :param weighted: If True, the probability of a text is weighted by its calculated 'probability_weight' attribute.
        :return: a random text index
        """
        if weighted is True:
            return random.choices(self.tidx, self.tcum)[0]
        else:
            return random.choice(self.tidx)

    def filter_text(self, text, sanitize_white_space=False, separate_punctuation=False, preserve_case=True):
        """ Filter a text.
        
        :param text: text to filter
        :param sanitize_white_space: If True, white space is replaced by a single space.
        :param separate_punctuation: If True, punctuation is separated from words.
        :param preserve_case: If True, the case of the text is preserved.
        :return: filtered text
        """
        if preserve_case is False:
            text = text.lower()
        punctuation = '!"#$%&\'()*+,-./:;<=>?@[\\]^_`{|}~'
        if separate_punctuation is True:
            for p in punctuation:
                text = text.replace(p, f' {p} ')
        if sanitize_white_space is True:
            text = text.replace('\n', ' ')
            text = text.replace('\t', ' ')
            text = text.replace('\r', ' ')
            text = text.replace('\f', ' ')
            text = text.replace('\v', ' ')
            to=""
            while to != text:
                to = text
                text = text.replace('  ', ' ')
        return text

    def get_random_sample(self, length, weighted=True):
        """ Return index idx and random sample of `length` chars from text[idx] the Text_Dataset.
        
        :param length: number of characters to return
        :param sanitize_white_space: If True, white space is replaced by a single space.
        :param separate_punctuation: If True, punctuation is separated from words.
        :param preserve_case: If True, the case of the text is preserved.
        :param weighted: If True, the probability of a text is weighted by its calculated 'probability_weight' attribute.
        :return: tuple (idx of text used for sampling, string of length `length` sampled from the Text_Dataset)
        """
        idx = self._get_random_text_index(weighted)
        text = self.text_list[idx]['text']
        if len(text) < length:
            sample = text
            while len(sample) < length:
                sample += ' '
        else:
            pos = random.randint(0, len(text) - length)
            sample = text[pos:pos+length]
        return (idx, sample)

    def _word_splitter(self, text):
        tokens=text.split()
        return tokens

    def init_tokenizer(self, tokenizer='word'):
        """ Initialize the tokenizer with the text_list.
        
        :param tokenizer: 'word' or 'char'
        """
        if tokenizer == 'word':
            self.w2i = {}
            self.i2w = {}
            self.w2i['<unk>'] = 0
            self.i2w[0] = '<unk>'
            self.w2i['<pad>'] = 1
            self.i2w[1] = '<pad>'
            self.w2i['<eos>'] = 2
            self.i2w[2] = '<eos>'
            self.w2i['<sos>'] = 3
            self.i2w[3] = '<sos>'
            for text in self.text_list:
                tokens = self._word_splitter(text['text'])
                for token in tokens:
                    if token not in self.w2i:
                        self.w2i[token] = len(self.w2i)
                        self.i2w[len(self.w2i)-1] = token
            self.word_tokenizer_init=True
        elif tokenizer == 'char':
            self.i2c = {}
            self.c2i = {}
            self.c2i['<unk>'] = 0
            self.i2c[0] = '<unk>'
            self.c2i['<pad>'] = 1
            self.i2c[1] = '<pad>'
            self.c2i['<eos>'] = 2
            self.i2c[2] = '<eos>'
            self.c2i['<sos>'] = 3
            self.i2c[3] = '<sos>'
            for text in self.text_list:
                tokens = list(text['text'])
                unique_chars = set(tokens)
                for c in unique_chars:
                    if c not in self.c2i:
                        ind = len(self.i2c)
                        self.c2i[c] = ind
                        self.i2c[ind] = c
            self.char_tokenizer_init=True
        else:
            self.log.error(f"Unknown tokenizer {tokenizer}")
            raise ValueError(f"Unknown tokenizer {tokenizer}")

    def tokenize(self, text, tokenizer='word'):
        """ Tokenize a text.
        
        :param text: text to tokenize
        :return: list of tokens """
        tokens = []
        if tokenizer == 'word':
            if self.word_tokenizer_init is False:
                self.init_tokenizer(tokenizer)
            tokens = self._word_splitter(text)
        elif tokenizer == 'char':
            if self.char_tokenizer_init is False:
                self.init_tokenizer(tokenizer)
            tokens = list(text)
        else:
            self.log.error(f"Unknown tokenizer {tokenizer}")
            raise ValueError(f"Unknown tokenizer {tokenizer}")
        return tokens

    def encode(self, text, tokenizer='word'):
        """ Encode a text.
        
        :param text: text to encode
        :return: list of encoded tokens """
        tokens = self.tokenize(text, tokenizer=tokenizer)
        if tokenizer == 'word':
            encoded = [self.w2i[token] if token in self.w2i else self.w2i['<unk>'] for token in tokens]
        elif tokenizer == 'char':
            encoded = [self.c2i[token] if token in self.c2i else self.c2i['<unk>'] for token in tokens]
        else:
            self.log.error(f"Unknown tokenizer {tokenizer}")
            raise ValueError(f"Unknown tokenizer {tokenizer}")
        return encoded

    def decode(self, encoded, tokenizer='word'):
        """ Decode a list of encoded tokens.
        
        :param encoded: list of encoded tokens
        :return: text """
        if tokenizer == 'word':
            decoded = [self.i2w[token]+'' if token in self.i2w else self.i2w['<unk>'] for token in encoded]
            decoded_text = ' '.join(decoded)
        elif tokenizer == 'char':
            decoded = [self.i2c[token] if token in self.i2c else self.i2c['<unk>'] for token in encoded]
            decoded_text = ''.join(decoded)
        else:
            self.log.error(f"Unknown tokenizer {tokenizer}")
            raise ValueError(f"Unknown tokenizer {tokenizer}")
        return decoded_text

    def get_random_char_tokenized_sample_pair(self, length):
        """ Get a random tokenized sample of the dataset.
        
        :param length: length of the sample
        :return: tuple (X, y) encoded sample
        """
        _, sample = self.get_random_sample(length+1)
        e_sample = self.encode(sample, tokenizer='char')
        X = e_sample[:-1]
        y = e_sample[1:]
        return X, y

    def init_getitem(self, sample_type='chargen', sample_length=80, content_stepping=10):
        """ Initialize the __getitem__ and __len__ methods.

        This method needs to be called before using len() or index-access of the dataset.

        This method determines how the dataset is partitioned into records, and what kind
        of encoding is returned on index-access.

        .. code-block:: python

            from ml_indie_tools.Text_Dataset import TextDataset
            tl = [{'author':'nobody', 'title':'some title', 'language':'english', 'text':'some text'},
                  {'author':'nobody', 'title':'some title 2', 'language':'english', 'text':'some more text'}]
            td = Text_Dataset(tl)
            td.init_getitem(sample_type='chargen', sample_length=4, content_stepping=2)
            print(len(td))
            print(td[0])
            # Output: 12 and ('some', 'ome ')

        :param sample_type: 'chargen': generate a pair of text or length sample_length, shifted by one letter, or 'chargen_encoded', same but encoded.
        :param sample_length: length of a sample
        :param content_stepping: number of characters to skip between each sample
        :return: on 'chargen[_encoded]': X, y [encoded] strings of sample_length, y shifted by one letter.
        """

        self.getitem_sample_type = sample_type
        self.getitem_sample_length = sample_length
        self.getitem_content_stepping = content_stepping
        leng=0
        rec=0
        if sample_type=='chargen' or sample_type=='chargen_encoded':
            for ind in range(0, len(self.text_list)):
                len_text = len(self.text_list[ind]['text'])
                rec_text = (len_text-content_stepping+1)//content_stepping + 1
                self.text_list[ind]['records']=rec_text
                leng += len_text
                rec += rec_text
            self.getitem_length = leng
            self.getitem_records = rec
            self.getitem_init = True
        else:
            self.getitem_length = 0
            self.getitem_records = 0
            print(f"init_getitem: unknown sample_type {sample_type}")

    def __len__(self):
        """ Get the length of the dataset.

        Note that this length depends on the initialization via :ref:`~Text_Dataset.Text_Dataset.init_getitem`.

        :return: length of the dataset (mode dependent) 
        """
        if self.getitem_init is False:
            print("init_getitem must be called before __len__")
            return None
        return self.getitem_records

    def _getitem_chargen(self, index):
        if index<0:
            if index<(-self.getitem_records):
                raise IndexError(f"index {index} out of range")
            else:
                index += self.getitem_records
        if index >= self.getitem_records:
            raise IndexError(f"index {index} out of range")
        cur_rec = 0
        for text in self.text_list:
            rec = text['records']
            if cur_rec+rec > index:
                rel_rec = index - cur_rec
                pos = rel_rec*self.getitem_content_stepping
                sample = text['text'][pos:pos+self.getitem_sample_length+1]
                while len(sample) < self.getitem_sample_length+1:
                    sample += ' '
                X_text = sample[:-1]
                y_text = sample[1:]
                return X_text, y_text
            cur_rec += rec
        print("Internal error in __getitem__")
        raise ValueError("Internal error in __getitem__")

    def __getitem__(self, index):
        """ Get a sample from the dataset.

        Format of the returned sample depends on :ref:`~Text_Dataset.Text_Dataset.init_getitem`.

        :param index: index of the sample
        :return:
        """
        if self.getitem_init is False:
            print("init_getitem must be called before __getitem__")
            raise ValueError("init_getitem must be called before __getitem__")
        if self.getitem_sample_type == 'chargen':
            return self._getitem_chargen(index)
        elif self.getitem_sample_type == 'chargen_encoded':
            X, y = self._getitem_chargen(index)
            X = self.encode(X, tokenizer='char')
            y = self.encode(y, tokenizer='char')
            return X, y
        else:
            self.log.error(f"Unknown getitem sample_type {self.getitem_sample_type}")
            raise ValueError(f"Unknown getitem sample_type {self.getitem_sample_type}")

    def _display_colored_html(self, textlist, dark_mode=False, display_ref_anchor=True, pre='', post=''):
        """ Internal function to display text and citation references in HTML. """
        bgcolorsWht = ['#d4e6e1', '#d8daef', '#ebdef0', '#eadbd8', '#e2d7d5', '#edebd0',
                    '#ecf3cf', '#d4efdf', '#d0ece7', '#d6eaf8', '#d4e6f1', '#d6dbdf',
                    '#f6ddcc', '#fae5d3', '#fdebd0', '#e5e8e8', '#eaeded', '#A9CCE3']
        bgcolorsDrk = ['#342621','#483a2f', '#3b4e20', '#2a3b48', '#324745', '#3d3b30',
                    '#3c235f', '#443f4f', '#403c37', '#463a28', '#443621', '#364b5f',
                    '#264d4c', '#2a3553', '#3d2b40', '#354838', '#3a3d4d', '#594C23']
        if dark_mode is False:
            bgcolors=bgcolorsWht
        else:
            bgcolors=bgcolorsDrk
        out = ''
        for txt, ind in textlist:
            txt = txt.replace('\n', '<br>')
            if ind == 0:
                out += txt
            else:
                if display_ref_anchor is True:
                    anchor="<sup>[" + str(ind) + "]</sup>"
                else:
                    anchor=""
                out += "<span style=\"background-color:"+bgcolors[ind % 16]+";\">" + \
                       txt + "</span>"+ anchor
        display(HTML(pre+out+post))

    def source_highlight(self, ref_txt, min_quote_size=10, dark_mode=False, display_ref_anchor=True):
        """ Analyse which parts of `ref_txt` are cited from the texts in the Text_Dataset.
        
        Note: this function requires a jupyter notebook in order to display HTML with markup.
        
        :param ref_txt: the reference text to be analysed for plagiarised parts
        :param min_quote_size: minimum size of a quote to be considered plagiarised
        :param dark_mode: if True, the background colors will be dark, otherwise white
        :param display_ref_anchor: if True, the reference text will be displayed with a reference anchor
        """
        ref_tx = ref_txt
        out = []
        qts = []
        txsrc = [("Sources: ", 0)]
        sc = False
        noquote = ''
        while len(ref_tx) > 0:  # search all library files for quote 'txt'
            mxQ = 0
            mxI = 0
            mxN = ''
            found = False
            for text in self.text_list:  # find longest quote in all texts
                p = min_quote_size
                if p <= len(ref_tx) and ref_tx[:p] in text['text']:
                    p = min_quote_size + 1
                    while p <= len(ref_tx) and ref_tx[:p] in text['text']:
                        p += 1
                    if p-1 > mxQ:
                        mxQ = p-1
                        mxI = text['index']
                        mxN = f"{text['author']}: {text['title']}"
                        found = True
            if found:  # save longest quote for colorizing
                if len(noquote) > 0:
                    out.append((noquote, 0))
                    noquote = ''
                out.append((ref_tx[:mxQ], mxI))
                ref_tx = ref_tx[mxQ:]
                if mxI not in qts:  # create a new reference, if first occurence
                    qts.append(mxI)
                    if sc:
                        txsrc.append((", ", 0))
                    sc = True
                    txsrc.append((mxN, mxI))
            else:
                noquote += ref_tx[0]
                ref_tx = ref_tx[1:]
        if len(noquote) > 0:
            out.append((noquote, 0))
            noquote = ''
        self._display_colored_html(out, dark_mode=dark_mode, display_ref_anchor=display_ref_anchor)
        if len(qts) > 0:  # print references, if there is at least one source
            self._display_colored_html(txsrc, dark_mode=dark_mode, display_ref_anchor=display_ref_anchor, pre="<small><p style=\"text-align:right;\">",
                                     post="</p></small>")



import sys
import os
import logging
import random
import numpy as np


class ALU_Dataset():
    """ Generate training data for all ALU operations

    The ALU takes two integers and applies one of the supported
    model_ops. E.g. `op1=123, op2=100, op='-' -> result 23`

    The net is supposed to learn to 'calculate' the results for
    arbitrary op1, op2 (positive integers, `0..2**bit_count - 1`) and 
    the twelve supported ops: 
    `["+", "-", "*", "/", "%", "AND", "OR", "XOR", ">", "<", "=", "!="]`

    :param bit_count: number of bits for each of the two operands, default 31 (mult uses 15 bits)
    :param pre_weight: if True, the model_dis will be reweighted to generate samples for 'difficult' ops
    """
    def __init__(self, bit_count=31, pre_weight=False):
        self.log = logging.getLogger("Datasets")
        self.model_ops = ["+", "-", "*", "/", "%",
                          "AND", "OR", "XOR", ">", "<", "=", "!="]
        self.model_is_boolean = [False, False, False, False, False,
                                 False, False, False, True, True, True, True]
        # Probabilites for creating a sample for each of the ops, (Will be
        # reweighted on checks to generate for samples for 'difficult' ops):
        self.model_dis = [10, 10, 10, 10, 10, 10,   10,  10,   10, 10, 10, 10]
        model_dis_w = [19, 12, 110, 15, 36, 10, 10, 10, 10, 10, 10, 10]
        self.model_funcs = [self._add_smpl, self._diff_smpl, self._mult_smpl,
                            self._div_smpl, self._mod_smpl, self._and_smpl,
                            self._bor_smpl, self._xor_smpl, self._greater_smpl,
                            self._lesser_smpl, self._eq_smpl, self._neq_smpl]
        self.bit_count = bit_count
        self.op_count = len(self.model_ops)
        if self.bit_count+1 > self.op_count:
            self.embedding_size = self.bit_count+1
        else:
            self.embedding_size = self.op_count
        self.pre_weight = pre_weight
        self.all_bits_one = 2**self.bit_count - 1
        self.true_vect = self.all_bits_one
        self.false_vect = 0
        self.input_size = (self.bit_count+1)*2+len(self.model_ops)
        self.output_size = 32
        if pre_weight is True:
            self.model_dis=model_dis_w

    @staticmethod
    def _int_to_binary_vect(num_int, num_bits=8):
        """ get a binary encoded vector of n of bit-lenght nm 

        :param num_int: integer to encoded
        :param num_bits: number of bits to use for positional_encoding
        :return: binary vector of length num_bits
        """
        num_vect = np.zeros(num_bits, dtype=np.float32)
        for i in range(0, num_bits):
            if num_int & (2**i) != 0:
                num_vect[i] = 1.0
        return num_vect

    @staticmethod
    def _int_to_onehot_vect(num_int, num_bits):
        """ get a one-hot encoded vector of n of bit-lenght nm 

        :param num_int: integer to encoded
        :param num_bits: number of bits to use for positional_encoding
        :return: one-hot vector of length num_bits
        """
        num_vect = np.zeros(num_bits, dtype=np.float32)
        num_vect[num_int] = 1.0
        return num_vect

    @staticmethod
    def _get_random_bits(bits):
        """ get bits random int 0...2**bits-1 

        :param bits: number of bits to uses
        :return: random int `0...2**bits-1`
        """
        return random.randint(0, 2**bits-1)

    def op_string_to_index(self, op_string):
        """ transform op_string (e.g. '+' -> 0) into corresponding index 

        :param op_string: string of op to transform
        :return: index of op_string
        """
        for i in range(0, len(self.model_ops)):
            if self.model_ops[i] == op_string:
                return i
        return -1

    def get_data_point(self, equal_distrib=False, valid_ops=None, vector=False, positional_encoding=False):
        """ Get a random example for on ALU operation for training 

        :param equal_distrib: if False, more 'difficult' ops will be generated more often.
        :param valid_ops: if not None, only the ops in valid_ops will be used
        :param vector: if True, the result will be returned as an embedded encoded vector
        :param positional_encoding: if True, the result will be returned as an embedded encoded vector with additional bits for positional positional_encoding

        """
        # result = -1
        op1 = self._get_random_bits(self.bit_count)
        op2 = self._get_random_bits(self.bit_count)
        if valid_ops is not None and len(valid_ops)==0:
            valid_ops=None
        if valid_ops is not None:
            if equal_distrib is False:
                self.log.warning("Op restriction via valid_ops forces equal_distrib=True")
                equal_distrib=True
            for op in valid_ops:
                if op not in self.model_ops:
                    self.log.warning(f'Cannot restrict valid_ops to {op}, unknown operation, ignoring all valid_ops')
                    valid_ops=None
                    break

        if equal_distrib or valid_ops is not None:
            if valid_ops is None:   
                op_index = random.randint(0, len(self.model_ops)-1)
            else:
                if len(valid_ops)==1:
                    op_index=0
                else:
                    op_index = random.randint(0, len(valid_ops)-1)
                op_index=self.model_ops.index(valid_ops[op_index])
        else: # make 'difficult' ops more present in training samples:
            rx = 0
            for md in self.model_dis:
                rx += md
            rrx = random.randint(0, rx)
            rx = 0
            op_index = 0
            for op_index in range(0, len(self.model_ops)):
                rx += self.model_dis[op_index]
                if rx > rrx:
                    break
        return self._encode_op(op1, op2, op_index, vector=vector, positional_suffix=positional_encoding)

    def _generatorenerator(self, samples=20000, equal_distrib=False, valid_ops=None):
        while True:
            x, Y = self.create_training_data(samples=samples, valid_ops=valid_ops, equal_distrib=equal_distrib, title=None)
            yield x, Y

    def _encode_op(self, op1, op2, op_index, vector=False, positional_suffix=False):
        """ turn two ints and operation into training data """
        op1, op2, result = self.model_funcs[op_index](op1, op2)
        if self.model_is_boolean[op_index] is True:
            if result==self.false_vect:

                str_result="True"
            else:
                str_result="undefined"
        else:
            str_result=result
        sym = f"{op1} {self.model_ops[op_index]} {op2} = {str_result}"
        if vector is True:
            if positional_suffix is True:
                sz=self.embedding_size+3
            else:
                sz=self.embedding_size
            v1=self._int_to_binary_vect(op1, num_bits=sz)
            v2=self._int_to_onehot_vect(op_index, num_bits=sz)
            v3=self._int_to_binary_vect(op2, num_bits=sz)
            if positional_suffix is True:
                v1[-3] = 1.0
                v2[-2] = 1.0
                v3[-1] = 1.0
            inp = np.array([v1,v2,v3], dtype=np.float32)
        else:
            inp = np.concatenate(
                [self._int_to_binary_vect(op1, num_bits=self.bit_count+1),
                self._int_to_onehot_vect(op_index, num_bits=len(self.model_ops)),
                self._int_to_binary_vect(op2, num_bits=self.bit_count+1)])

        oup = self._int_to_binary_vect(result, num_bits=self.output_size)
        return inp, oup, result, op_index, sym

    @staticmethod
    def _add_smpl(op1, op2):
        """ addition training example """
        result = op1+op2
        return op1, op2, result

    @staticmethod
    def _diff_smpl(op1, op2):
        """ subtraction training example """
        if op2 > op1:
            op2, op1 = op1, op2
        result = op1-op2
        return op1, op2, result

    def _mult_smpl(self, op1, op2):
        """ multiplication training example """
        modul = 2**(self.bit_count//2) - 1
        op1 = op1 % modul
        op2 = op2 % modul
        result = op1*op2
        return op1, op2, result

    def _div_smpl(self, op1, op2):
        """ integer division training example """
        while op2 == 0:
            op2 = self._get_random_bits(self.bit_count)
        if op1 < op2 and random.randint(0, 2) != 0:
            if op1 != 0:
                op1, op2 = op2, op1
        result = op1//op2
        return op1, op2, result

    def _mod_smpl(self, op1, op2):
        """ modulo (remainder) training example """
        while op2 == 0:
            op2 = self._get_random_bits(self.bit_count)
        if op1 < op2 and random.randint(0, 2) != 0:
            if op1 != 0:
                op1, op2 = op2, op1
        result = op1 % op2
        return op1, op2, result

    @staticmethod
    def _and_smpl(op1, op2):
        """ bitwise AND training example """
        result = op1 & op2
        return op1, op2, result

    @staticmethod
    def _bor_smpl(op1, op2):
        """ bitwise OR training example """
        result = op1 | op2
        return op1, op2, result

    @staticmethod
    def _xor_smpl(op1, op2):
        """ bitwise XOR training example """
        result = op1 ^ op2
        return op1, op2, result

    def _greater_smpl(self, op1, op2):
        """ integer comparisation > training example """
        if op1 > op2:
            result = self.true_vect
        else:
            result = self.false_vect
        return op1, op2, result

    def _lesser_smpl(self, op1, op2):
        """ integer comparisation < training example """
        if op1 < op2:
            result = self.true_vect
        else:
            result = self.false_vect
        return op1, op2, result

    def _eq_smpl(self, op1, op2):
        """ integer comparisation == training example """
        if random.randint(0, 1) == 0:  # create more cases
            op2 = op1
        if op1 == op2:
            result = self.true_vect
        else:
            result = self.false_vect
        return op1, op2, result

    def _neq_smpl(self, op1, op2):
        """ integer comparisation != training example """
        if random.randint(0, 1) == 0:  # create more cases
            op2 = op1
        if op1 != op2:
            result = self.true_vect
        else:
            result = self.false_vect
        return op1, op2, result

    def create_data_point(self, op1, op2, op_string, vector=False, positional_suffix=False):
        """ create training data from given ints op1, op2 and op_string """
        op_index = self.op_string_to_index(op_string)
        if op_index == -1:
            print(f"Invalid operation {op_string}")
            return np.array([]), np.array([]), -1, -1, None
        return self._encode_op(op1, op2, op_index, vector, positional_suffix)

    def create_training_data(self, samples=10000, valid_ops=None, equal_distrib=False, verbose=True, title=None):
        """ create a number of training samples """
        x, y, _, _, _ = self.get_data_point()
        dpx = np.zeros((samples, len(x)), dtype=np.float32)
        dpy = np.zeros((samples, len(y)), dtype=np.float32)
        if verbose is True:
            if title is None:
                print(f"Creating {samples} data points (. = 1000 progress)")
            else:
                print(f"{title}: Creating {samples} data points (. = 1000 progress)")

        for i in range(0, samples):
            if verbose is True:
                if i%100000 == 0:
                    print(f"{i:>10} ", end="")
            if (i+1) % 1000 == 0:
                if verbose is True:
                    print(".", end="")
                    sys.stdout.flush()
                    if (i+1) % 100000 == 0:
                        print()
            if valid_ops is None:
                x, y, _, _, _ = self.get_data_point(equal_distrib=equal_distrib)
            else:
                x, y, _, _, _ = self.get_data_point(
                    equal_distrib=True, valid_ops=valid_ops)
            dpx[i, :] = x
            dpy[i, :] = y
        if verbose is True:
            print()
        return dpx, dpy

    def create_vector_training_data(self, samples=10000, valid_ops=None, equal_distrib=False, verbose=True, title=None, positional_encoding=True):
        """ create a number of training samples """
        x, y, _, _, _ = self.get_data_point()
        if positional_encoding is True:
            sz=self.embedding_size+3
        else:
            sz=self.embedding_size
        dpx = np.zeros((samples, 3, sz), dtype=np.float32)
        dpy = np.zeros((samples, len(y)), dtype=np.float32)
        if verbose is True:
            if title is None:
                print(f"Creating {samples} data points (. = 1000 progress)")
            else:
                print(f"{title}: Creating {samples} data points (. = 1000 progress)")

        for i in range(0, samples):
            if verbose is True:
                if i%100000 == 0:
                    print(f"{i:>10} ", end="")
            if (i+1) % 1000 == 0:
                if verbose is True:
                    print(".", end="")
                    sys.stdout.flush()
                    if (i+1) % 100000 == 0:
                        print()
            if valid_ops is None:
                x, y, _, _, _ = self.get_data_point(equal_distrib=equal_distrib, vector=True, positional_encoding=positional_encoding)
            else:
                x, y, _, _, _ = self.get_data_point(
                    equal_distrib=True, valid_ops=valid_ops, vector=True, positional_encoding=positional_encoding)
            dpx[i, :, :] = x
            dpy[i, :] = y
        if verbose is True:
            print()
        return dpx, dpy

    if 'tensorflow' in sys.modules:
        def create_dataset(self, samples=10000, batch_size=2000, vector=False, positional_encoding=True, is_training=True, valid_ops=None, name=None, cache_path=None, use_cache=True, regenerate_cached_data=False, for_tpu=False):
            import tensorflow as tf
            is_loaded=False
            if use_cache is True and cache_path is None:
                print("can't use cache if no cache_path is given, disabling cache!")
                use_cache = False
            if use_cache is True:
                if valid_ops is not None:
                    infix='_'
                    for vo in valid_ops:
                        if vo=='*': # Prevent poison-filenames
                            vo="MULT"
                        if vo=='/':
                            vo="DIV"
                        if vo=='%':
                            vo='MOD'
                        if vo=='<':
                            vo='LT'
                        if vo=='>':
                            vo='GT'
                        if vo=='=':
                            vo='EQ'
                        if vo=='!=':
                            vo='NE'
                        infix+=vo
                else:
                    infix=""
                if vector is True:
                    infix+="_VECT"
                if positional_encoding is True:
                    infix+="_PE"
                cache_file_x=os.path.join(cache_path, f"{name}_{infix}_{self.bit_count}_{samples}_x.npy")
                cache_file_Y=os.path.join(cache_path, f"{name}_{infix}_{self.bit_count}_{samples}_Y.npy")
            else:
                cache_file_x=None
                cache_file_Y=None
            if use_cache is True  and regenerate_cached_data is False and os.path.exists(cache_file_x) and os.path.exists(cache_file_Y):
                try:
                    x = np.load(cache_file_x, allow_pickle=True)
                    Y = np.load(cache_file_Y, allow_pickle=True)
                    if len(x)==samples:
                        is_loaded=True
                        self.log.debug(f"Data {name} loaded from cache")
                    else:
                        self.log.info(f"Sample count has changed from {len(x)} to {samples}, regenerating {name} data...")
                except Exception as e:
                    self.log.error(f"Something went wrong when loading {cache_file_x}, {cache_file_Y}: {e}")
            if is_loaded is False:
                if vector is True:
                    x, Y = self.create_vector_training_data(
                        samples=samples, valid_ops=valid_ops, title=name, positional_encoding=positional_encoding)
                else:
                    x, Y = self.create_training_data(samples=samples, valid_ops=valid_ops, title=name)
                if use_cache is True:
                    self.log.debug(f"Writing data-cache {cache_file_x}, {cache_file_Y}...")
                    np.save(cache_file_x, x, allow_pickle=True)
                    np.save(cache_file_Y, Y, allow_pickle=True)
                    self.log.debug("x, Y, done.")
            shuffle_buffer=10000
            dataset=tf.data.Dataset.from_tensor_slices((x, Y)).cache()
            if is_training is True:
                dataset=dataset.shuffle(shuffle_buffer, reshuffle_each_iteration=True)
                if for_tpu is True:
                    dataset=dataset.repeat() # Mandatory for Keras TPU for now
            dataset=dataset.batch(batch_size, drop_remainder=True) # drop_remainder is important on TPU, batch size must be fixed
            dataset=dataset.prefetch(-1) # fetch next batches while training on the current one (-1: autotune prefetch buffer size)
            return dataset

        def create_dataset_from_generator(self, valid_ops=None):
            import tensorflow as tf
            dataset=tf.data.Dataset.from_generator(
                self._generator,
                output_signature=(
                        tf.TensorSpec(shape=(None,self.input_size), dtype=np.float32),
                        tf.TensorSpec(shape=(None,self.output_size), dtype=np.float32))
                )
            return dataset

        def get_datasets(self, pre_weight=True, samples=100000, validation_samples=10000, batch_size=2000, vector=False, positional_encoding=True, valid_ops=None, cache_path=None, use_cache=True, regenerate_cached_data=False, for_tpu=False):
            train = self.create_dataset(samples=samples, batch_size=batch_size, is_training=True, vector=vector, positional_encoding=positional_encoding, valid_ops=valid_ops,
                                            name="train",cache_path=cache_path, use_cache=use_cache, regenerate_cached_data=regenerate_cached_data, for_tpu=for_tpu)
            val = self.create_dataset(samples=validation_samples, batch_size=batch_size, vector=vector, positional_encoding=positional_encoding, is_training=False, valid_ops=valid_ops,
                                        name="validation",cache_path=cache_path, use_cache=use_cache, regenerate_cached_data=regenerate_cached_data)
            return train, val
    else:
        def get_datasets(self, pre_weight=True, samples=100000, validation_samples=10000, batch_size=2000, vector=False, positional_encoding=True, valid_ops=None, cache_path=None, use_cache=True, regenerate_cached_data=False):
            print("No tensorflow, no datasets!")
            return None, None

    def decode_results(self, result_int_vects):
        """ take an array of 32-float results from neural net and convert to ints """
        result_vect_ints = []
        for vect in result_int_vects:
            if (len(vect) != self.output_size):
                print(f"Ignoring unexpected vector of length {len(vect)}")
            else:
                int_result = 0
                for i in range(0, self.output_size):
                    if vect[i] > 0.5:
                        int_result += 2**i
                result_vect_ints.append(int_result)
        return result_vect_ints

    def check_results(self, model, samples=1000, vector=False, positional_encoding=True, valid_ops=None, verbose=False):
        """ Run a number of tests on trained model """
        ok = 0
        err = 0
        operr = [0]*len(self.model_ops)
        opok = [0]*len(self.model_ops)
        for _ in range(0, samples):
            x, _, z, op, s = self.get_data_point(
                equal_distrib=True, vector=vector, positional_encoding=positional_encoding, valid_ops=valid_ops)
            res = self.decode_results(model.predict(np.array([x])))
            if res[0] == z:
                ok += 1
                opok[op] += 1
                r = "OK"
            else:
                err += 1
                operr[op] += 1
                r = "Error"
            if verbose is True:
                if self.model_is_boolean[op] is True:
                    if res[0]==self.false_vect:
                        str_result="False"
                    elif res[0]==self.true_vect:
                        str_result="True"
                    else:
                        str_result="undefined"
                else:
                    str_result=res[0]
                if res[0]==z:
                    print(f"{s} == {str_result}: {r}")
                else:
                    print(f"{s} != {str_result}: {r}")
                    if self.model_is_boolean[op] is False:
                        print(bin(res[0]))
                        print(bin(z))
        opsum = ok+err
        if opsum == 0:
            opsum = 1
        print(f"Ok: {ok}, Error: {err} -> {ok/opsum*100.0}%")
        print("")
        for i in range(0, len(self.model_ops)):
            opsumi = opok[i]+operr[i]
            if opsumi == 0:
                continue
            # modify the distribution of training-data generated to favour
            # ops with bad test results, so that more training data is
            # generated on difficult cases:
            self.model_dis[i] = int(operr[i]/opsumi*100)+10
            print(
                f"OP{self.model_ops[i]}: Ok: {opok[i]}, Error: {operr[i]}", end="")
            print(f" -> {opok[i]/opsumi*100.0}%")
        if valid_ops == None:
            print("Change probability for ops in new training data:")
            print(f"Ops:     {self.model_ops}")
            print(f"Weights: {self.model_dis}")
        return ok/opsum

